/*
 * This file is part of SubLink.
 * Copyright (C) 2009  Jeferson Hultmann <hultmann@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

var EXPORTED_SYMBOLS = ["buildResolveMenu"];

Components.utils.import("resource://sublink-migreme-modules/util.js");

function buildResolveMenu(mainMenu, cmdData, isUrlbar) {
  util.log("buildResolveMenu");

  if (util.isRedirUri(cmdData.uri) === false) {
    throw util.logStack("buildResolveMenu 0 " + cmdData.uri.spec);
  }

  var doc = mainMenu.ownerDocument;
  var menuParent = doc.getElementById("sublink-migreme-menu-resolve");
  if (menuParent) {
    throw util.logStack("buildResolveMenu 1 " + menuParent.tagName);
  }

  var sepItem = mainMenu.insertBefore(doc.createElement("menuseparator"), mainMenu.firstChild);
  menuParent = mainMenu.insertBefore(doc.createElement("menu"), mainMenu.firstChild);
  menuParent.setAttribute("id", "sublink-migreme-menu-resolve");

  util.addClass(sepItem, "sublink-migreme-node");
  util.addClass(menuParent, "sublink-migreme-node");
  util.addClass(menuParent, "menu-iconic");

  if (cmdData.typed) {
    var copy = isUrlbar || cmdData.node.readOnly;
    menuParent.value = copy ? "popup-input-readonly" : "popup-input-default";
  } else {
    menuParent.value = "popup-link";
  }

  menuParent.setAttribute("disabled", "true");
  menuParent.setAttribute("label", cmdData.uri.spec);
  var online = util.isOffline() === false;
  if (online) {
    menuParent.setAttribute("image", "chrome://browser/skin/places/searching_16.png");
  }

  var redirPath = redirCache().get(cmdData.uri.spec);
  if (redirPath !== null) {
    util.log("getUserData found " + redirPath);
    buildRedirPopup(mainMenu.ownerDocument, cmdData, redirPath);

  } else {
    // TODO long url pode estar na lib, mas ainda eh necessario mapear o redir
    //if (redirStorage.loadShortUrls(longUri).hasOwnProperty(cmdData.uri.spec)) {
    if (online) {
      findRedirCore(mainMenu, cmdData);
    }
  }

  util.log("/buildResolveMenu");
}


function onPopupHidden(evt) {
  util.log("onPopupHidden*");
  var popup = evt.originalTarget;
  popup.removeEventListener("popuphidden", onPopupHidden, false);
  evt.stopPropagation();
}


/*
<INPUT EDIT>    LONG_URL
  *  Compactar com tinyurl: "http://longurl.com/..."

<INPUT READONLY>LONG_URL
  *  Copiar como tinyurl.com "http://longurl.com/..."


<INPUT EDIT>    SHORT_URL
   ! http://longurl.com/...>  Descompactar           [input=>replace]
                              Converter para tinyurl [input=>convert_replace]  (p/ SHORT_URL!=tinyurl)
<INPUT READONLY>SHORT_URL
   ! http://longurl.com/...>  Copiar                 [input=>copy]
                              Copiar como tinyurl    [input=>convert_copy]      (p/ SHORT_URL!=tinyurl)

<A>             SHORT_URL
  ! http://longurl.com/...>   Copiar                 [link=>copy]
                              Copiar como tinyurl    [link=>convert_copy]       (p/ SHORT_URL!=tinyurl)
*/

//menuItem.style.width='40em';//436px
//menuItem.style.width='30em';//326px
//menuItem.style.width='28em';
/*
util.log(doc.defaultView.getComputedStyle(menuItem, '').getPropertyValue("width"));
util.log(menuItem.ownerDocument.defaultView.getComputedStyle(menuItem, '').getPropertyValue("width"));
util.log(menuItem.ownerDocument.defaultView.getComputedStyle(menuItem, '').getPropertyCSSValue("width").cssText);
util.log(menuItem.ownerDocument.defaultView.getComputedStyle(menuItem, '').width);
*/

function findRedirCore(appMenu, cmdData) {

  var doc = appMenu.ownerDocument;
  var browser = doc.defaultView.getBrowser().selectedBrowser;

  var m_redirPath = [{
    delta: -1,
    responseStatusCode: -1,
    longUri: cmdData.uri
  }];


  function onRedirResolved(info) {
    var isInfiniteRedir = false;
    for (var idx = 0; idx < m_redirPath.length; idx++) {
      if (info.longUri.equals(m_redirPath[idx].longUri)) {
        isInfiniteRedir = true;
        break;
      }
    }

    m_redirPath.push(info);
    isInfiniteRedir = isInfiniteRedir || m_redirPath.__count__ > 16; // WTF m_redirPath nao eh array?

    if (util.isRedirUri(info.longUri) && isInfiniteRedir === false) {
      // nested url
      findRedirLocation(info.longUri, browser, onRedirResolved, onError);
      return;
    }

    if (hasRedirPathError(m_redirPath) === false) {
      redirCache().set(cmdData.uri.spec, m_redirPath);
    }
    buildRedirPopup(doc, cmdData, m_redirPath);
  }


  function onError(errData) {
    util.log("FAIL " + errData.toSource());
    if (appMenu.state !== "open") {
      return;
    }

    function cmdShow(evt) {
      Components.utils.import("resource://sublink-migreme-modules/err.js");
      notificationError(errData, browser);
    }

    var menu = doc.getElementById("sublink-migreme-menu-resolve");
    menu.setAttribute("image", "chrome://global/skin/icons/error-16.png");
    menu.removeAttribute("disabled");
    menu.addEventListener("popuphidden", onPopupHidden, false);


    var item = menu.appendChild(doc.createElement("menupopup"))
                   .appendChild(doc.createElement("menuitem"));
    item.setAttribute("label", util.getText("context-resolve.error.label"));
    item.setAttribute("accesskey", util.getText("context-resolve.error.accesskey"));
    item.addEventListener("command", cmdShow, false);
  }

  findRedirLocation(cmdData.uri, browser, onRedirResolved, onError);
}


/*
  var timeoutId = {};
  timeoutId.value = doc.defaultView.setTimeout(function() {
    configLongUrlFoundItems(mainMenu, cmdData, menuItem, copy, timeoutId);
  }, 0);
  menuItem.setAttribute("sublink-migreme-timeoutID", timeoutId.value);



  if (menuItem.getAttribute("sublink-migreme-timeoutID") !== timeoutId.value.toString()) {
    return;
  }
*/

function buildRedirPopup(doc, cmdData, redirPath) {
  var menu = doc.getElementById("sublink-migreme-menu-resolve");
  if (menu === null) {
    // abort: menu is gone

    try {
      Cc["@mozilla.org/sound;1"]
        .createInstance(Ci.nsISound)
        .playEventSound(Ci.nsISound.EVENT_NEW_MAIL_RECEIVED);
    } catch (ex) {
      util.log(ex);
    }
    return;
  }



  if (menu.hasAttribute("disabled") === false) {
    // abort: simultaneous requests
    return;
  }


  var info = redirPath[redirPath.length - 1];
  var fav = getFavicon(info.longUri).spec;


  menu.setAttribute("image", fav);
  menu.addEventListener("popuphidden", onPopupHidden, false);
  menu.removeAttribute("disabled");
  menu.setAttribute("label", info.longUri.prePath);
  switch (info.longUri.scheme) {
    case "http":
    case "https":
      menu.setAttribute("accesskey", "h");
      break;
    case "ftp":
      menu.setAttribute("accesskey", "f");
      break;
  }


  var menuNew = menu.appendChild(doc.createElement("menupopup"));
  var pathMenuItem = menuNew.appendChild(doc.createElement("menuitem"));

  menuNew.appendChild(doc.createElement("menuseparator"));

  var copyMenuItem = menuNew.appendChild(doc.createElement("menuitem"));
  var convertMenuItem = menuNew.appendChild(doc.createElement("menuitem"));
  var libraryItem = menuNew.appendChild(doc.createElement("menuitem"));

  menuNew.appendChild(doc.createElement("menuseparator"));

  buildRedirPath(menuNew, redirPath);


  var cmdHandler = new OnLongUriCmds(cmdData,
                                     info.longUri,
                                     doc.defaultView.getBrowser().selectedBrowser);

  configMenuItemPath(pathMenuItem, fav, redirPath);
  configMenuItemAction(copyMenuItem, cmdHandler, menu.value);


  if (cmdData.uri.host === "migre.me") {//basedomain
    configMenuItemRemember(libraryItem, cmdHandler, info.longUri, cmdData.uri);
    convertMenuItem.hidden = true;
  } else {
    configMenuItemConvert(convertMenuItem, cmdHandler, menu.value);
    libraryItem.hidden = true;
  }

}


function findProperty(redirPath, name) {
  for (var idx = redirPath.length - 1; idx > -1; idx--) {
    var info = redirPath[idx];
    if (info.hasOwnProperty(name) && (info[name] !== null)) {
      return info[name];
    }
  }
  return null;
}


function configMenuItemPath(menuItem, fav, redirPath) {

  var info = redirPath[redirPath.length - 1];
  var longUri = info.longUri;
  var title = util.pageTitle(longUri);

  if (!title) {
    title = findProperty(redirPath, "htmlTitle");
    //if (title) {
    //  util.log(title + "\n" + decodeURIComponent(title));
    //  title = decodeURIComponent(title); // workaround==>migre.me bug
    //}
  }

  var ct = findProperty(redirPath, "contentType");
  if (!title && ct) {
    if (ct.indexOf("image/") === 0) {
      var ct2 = ct.split("/", 2);
      title = util.getText("context-resolve.title.content-type.image.label", ct2[1].toUpperCase());
    } else if (ct.indexOf("text/") !== 0) {
      menuItem.setAttribute("description", ct);
    }
  }


  function onclick(evt) {
    util.openPage2(evt, longUri.spec, true);
  }


  if (!title) {
    // moz-anno:favicon:http://blog.mozilla.com/favicon.ico
    menuItem.setAttribute("image", fav);
    util.addClass(menuItem, "menuitem-iconic");
    menuItem.setAttribute("tooltiptext", longUri.spec);

  } else {
    var titleItem = menuItem.parentNode.insertBefore(
                      menuItem.ownerDocument.createElement("menuitem"),
                      menuItem);

    titleItem.setAttribute("image", fav);
    util.addClass(titleItem, "menuitem-iconic");

    var ttt = title + "\n" + longUri.spec;
    titleItem.setAttribute("tooltiptext", ttt);
    menuItem.setAttribute("tooltiptext", ttt);

    titleItem.setAttribute("label", title);
    //titleItem.setAttribute("crop", "center");
    titleItem.setAttribute("default", "true");
    setStatusBar(titleItem, longUri.spec);
    //titleItem.setAttribute("disabled", "true");


    titleItem.addEventListener("command", onclick, false);
    titleItem.addEventListener("click", onclick, false);
  }


  menuItem.addEventListener("command", onclick, false);
  menuItem.addEventListener("click", onclick, false);

  menuItem.setAttribute("label", longUri.path);
  menuItem.setAttribute("crop", "center");
  menuItem.setAttribute("accesskey", "/");
  setStatusBar(menuItem, longUri.spec);


  /*
  if (!title) {
    menuItem.setAttribute("tooltiptext", longUri.spec);
    menuItem.setAttribute("label", longUri.path);
    menuItem.setAttribute("crop", "center");

  } else {
    menuItem.setAttribute("description", title);
    menuItem.setAttribute("tooltiptext", title + "\n" + longUri.spec);
    if (longUri.path.length > 40) {
      // ensure "description" will appear
      menuItem.setAttribute("label", longUri.path.substr(0, 36) + "\u2026");
    } else {
      menuItem.setAttribute("label", longUri.path);
    }
  }
  */


  // http://developer.yahoo.com/rss/#terms
  // http://api.search.yahoo.com/WebSearchService/rss/webSearch.xml?appid=yahoosearchwebrss&query=url%3Ahttp%3A%2F%2Fblogs.msdn.com%2Fbrada%2Farticles%2F371015.aspx&adult_ok=1

  // http://code.google.com/intl/pt-BR/apis/ajaxsearch/documentation/#fonje
  // http://ajax.googleapis.com/ajax/services/search/web?v=1.0&q=info:http%3A%2F%2Fblogs.msdn.com%2Fbrada%2Farticles%2F371015.aspx
}


function configMenuItemRemember(menuItem, cmdHandler, longUri, redir) {
  menuItem.addEventListener("command", cmdHandler, false);
  menuItem.setAttribute("label", util.getText("context-resolve.remember.label"));
  menuItem.setAttribute("accesskey", util.getText("context-resolve.remember.accesskey"));
  menuItem.value = "add_library";

  Components.utils.import("resource://sublink-migreme-modules/redir-storage.js");
  var redirs = redirStorage.loadShortUrls(longUri);
  if (redirs.hasOwnProperty(redir.spec)) {
    menuItem.setAttribute("disabled", "true");
  }
}


function configMenuItemAction(menuItem, cmdHandler, menuType) {
  switch (menuType) {
    case "popup-link":
      //menuItem.value = "copy";
      //menuItem.setAttribute("label", util.getText("context.link.loaded.label"));
      //menuItem.setAttribute("accesskey", util.getText("context.link.loaded.accesskey"));
      //break;

    case "popup-input-readonly":
      menuItem.value = "copy";
      menuItem.setAttribute("label", util.getText("context-resolve.copy.label"));
      menuItem.setAttribute("accesskey", util.getText("context-resolve.copy.accesskey"));
      break;

    case "popup-input-default":
      menuItem.value = "replace";
      menuItem.setAttribute("label", util.getText("context-resolve.replace.label"));
      menuItem.setAttribute("accesskey", util.getText("context-resolve.replace.accesskey"));
      break;

    default:
      throw util.logStack(menuType);
  }

  menuItem.addEventListener("command", cmdHandler, false);
}


function configMenuItemConvert(menuItem, cmdHandler, menuType) {
  switch (menuType) {
    case "popup-link":
      //menuItem.value = "convert_copy";
      //menuItem.setAttribute("label", util.getText("context.link.convertCopy.label", "migre.me"));
      //menuItem.setAttribute("accesskey", util.getText("context.link.convertCopy.accesskey"));
      //menuItem.setAttribute("image", "http://migre.me/migreme.ico");
      //util.addClass(menuItem, "menuitem-iconic");
      //break;

    case "popup-input-readonly":
      menuItem.value = "convert_copy";
      menuItem.setAttribute("label", util.getText("context-resolve.convert.copy.label", "migre.me"));
      menuItem.setAttribute("accesskey", util.getText("context-resolve.convert.copy.accesskey"));
      menuItem.setAttribute("image", "http://migre.me/migreme.ico");
      util.addClass(menuItem, "menuitem-iconic");
      break;

    case "popup-input-default":
      menuItem.value = "convert_replace";
      // TODO ver se tem salvo
      menuItem.setAttribute("label", util.getText("context-resolve.convert.replace.label", "migre.me"));
      menuItem.setAttribute("accesskey", util.getText("context-resolve.convert.replace.accesskey"));
      menuItem.setAttribute("image", "http://migre.me/migreme.ico");
      util.addClass(menuItem, "menuitem-iconic");
      break;

    default:
      throw util.logStack(menuType);
  }

  menuItem.addEventListener("command", cmdHandler, false);
}


function hasRedirPathError(redirPath) {
  for (var idx = 0, len = redirPath.length - 1; idx < len ; idx++) {
    var httpStatus = redirPath[idx + 1].responseStatusCode;
    if (httpStatus === -200) {
      // redirectionLimit error
      return true;
    }
  }
  return false;
}


function buildRedirPath(menu, redirPath) {
  var doc = menu.ownerDocument;

  for (var idx = 0, len = redirPath.length - 1; idx < len ; idx++) {
    var nestedItem = menu.appendChild(doc.createElement("menuitem"));
    var myUri = redirPath[idx].longUri;

    nestedItem.setAttribute("label", myUri.spec);
    //if (redirPath[idx].delta > -1) {
    //  nestedItem.setAttribute("tooltiptext", redirPath[idx].delta + "ms"); // TODO
    //}
    var httpStatus = redirPath[idx + 1].responseStatusCode;

    if (httpStatus > 0 && httpStatus !== 301) {
      var desc = util.getText("context-resolve.http-status.description", httpStatus);
      nestedItem.setAttribute("description", desc);
    }

    if (httpStatus === -200) {
      // redirectionLimit error
      var errorItem = menu.appendChild(doc.createElement("menuitem"));
      util.addClass(errorItem, "menuitem-iconic");
      errorItem.setAttribute("image", "chrome://global/skin/icons/warning-16.png");
      errorItem.setAttribute("label", util.getText("context-resolve.pathError"));
      errorItem.setAttribute("disabled", "true");
    }

    var info = util.getInfoUrl(myUri);
    if (info === null) {
      nestedItem.setAttribute("disabled", "true");
      continue;
    }

    nestedItem.addEventListener("command", onRedirInfoCommand, false);
    nestedItem.addEventListener("click", onRedirInfoCommand, false);
    nestedItem.setAttribute("image", "chrome://global/skin/icons/information-16.png");
    nestedItem.value = info;
    util.addClass(nestedItem, "menuitem-iconic");
    setStatusBar(nestedItem, info);
  }


  //if for (var idx = 0, len = redirPath.length - 1; idx < len ; idx++) {

}


function onRedirInfoCommand(evt) {
  var item = evt.target;
  var popup = item.parentNode;
  var items = popup.querySelectorAll("menuitem[image='chrome://global/skin/icons/information-16.png']");
  var close = items.length < 2;
  var url = item.value;
  util.openPage2(evt, url, close);
}


function findRedirLocation(uri, browserOrigin, onSuccess, error) {

  if (util.isRedirUri(uri) === false) {
    throw util.logStack(uri.host);
  }

  var req;
  var baseDomain = Cc["@mozilla.org/network/effective-tld-service;1"]
                    .getService(Ci.nsIEffectiveTLDService)
                    .getBaseDomain(uri);


  switch (baseDomain) {
    case "migre.me":
      Components.utils.import("resource://sublink-migreme-modules/api-migreme.js");
      req = new AsyncRequest();
      break;
    default:
      if (util.isBitLy(baseDomain)) {
        Components.utils.import("resource://sublink-migreme-modules/resolve-redir.js");
        req = new ResolveRedir();
      } else {
        Components.utils.import("resource://sublink-migreme-modules/resolve-redir.js");
        req = new ResolveRedir();
      }
      break;
  }


  req.onSuccess = onSuccess;
  req.onError = error;

  //var start = new Date().getTime();
  var chromeWnd = browserOrigin.ownerDocument.defaultView;
  chromeWnd.setCursor("progress");

  req.onComplete = function() {
    chromeWnd.setCursor("auto");
    //util.log("resolve redir connection: " + (new Date().getTime() - start) + "ms " + uri.spec);
  };

  req.findRedir(uri);
}


//copy / copy as / remember
function OnLongUriCmds(c, l, b) {
  this.cmdData = c;
  this.longUri = l;
  this.browserOrigin = b;
}


OnLongUriCmds.prototype = {

  handleEvent: function(evt) {
    var menuItem = evt.target;
    var cmd = menuItem.value;
    switch (cmd) {
      case "convert_copy":
      case "convert_replace":
      case "copy":
      case "replace":
      case "add_library":
        this[cmd]();
        break;
      default:
        throw util.logStack(cmd);
    }
  },

  add_library: function() {
    util.log("add_library " + this.longUri.spec + "\n" + this.cmdData.toSource());
    if (isExtRedir(this.cmdData.uri) === false) {
      throw util.logStack(this.cmdData.uri);
    }

    Components.utils.import("resource://sublink-migreme-modules/redir-storage.js");
    redirStorage.saveShortUrl(this.longUri, this.cmdData.uri);
  },

  copy: function() {
    util.toClipboard(this.longUri.spec);
  },

  replace: function() {
    Components.utils.import("resource://sublink-migreme-modules/run-input.js");
    PUBLIC_replaceCore(this.cmdData, this.longUri.spec, this.browserOrigin);
  },

  convert_copy: function() {
    function onSuccess(info) {
      util.toClipboard(info.shortUri.spec);
    }
    util.toClipboard("");
    util.toTinyUrlRequest(this.longUri, this.browserOrigin, onSuccess);
  },

  convert_replace: function() {
    var b = this.browserOrigin;
    var me = this;
    function onSuccess(info) {
      Components.utils.import("resource://sublink-migreme-modules/run-input.js");
      PUBLIC_replaceCore(me.cmdData, info.shortUri.spec, b);
    }

    util.toTinyUrlRequest(this.longUri, b, onSuccess);
  }

}


function setStatusBar(item, text) {
  item.setAttribute("sublink-migreme-statusbartext", text);

  item.addEventListener("DOMMenuItemActive", function(evt) {
    //var menuitem = evt.target;
    setStatusBarCore(this, this.getAttribute("sublink-migreme-statusbartext"));
    //util.log(this.tagName);
  }, false);

  item.addEventListener("DOMMenuItemInactive", function(evt) {
    //var menuitem = evt.target;
    setStatusBarCore(this, ""); //this=menuitem
  }, false);
}

function setStatusBarCore(item, text) {
  item.ownerDocument.defaultView.XULBrowserWindow.setOverLink(text);
}


function getFavicon(uri) {
  return Cc["@mozilla.org/browser/favicon-service;1"]
          .getService(Ci.nsIFaviconService)
          .getFaviconImageForPage(uri);
}

function isExtRedir(uri) {
  return uri.host === "migre.me";
}


var _cache = null;
function redirCache() {
  if (this._cache === null) {
    this._cache = new ObjCache(4);
  }
  return this._cache;
}


function ObjCache(len) {
  this._keys = new Array(len);
  this._values = new Array(len);
  this._nextIndex = 0;
}

ObjCache.prototype = {

  get: function(key) {
    var idx = this._keys.indexOf(key);
    if (idx === -1) {
      return null;
    }
    var val = this._values[idx];
    this._set(key, val);
    return val;
  },

  set: function(key, val) {
    var idx = this._keys.indexOf(key);

    if (idx > -1) {
      this._keys[idx] = null;
      this._values[idx] = null;
    }

    this._set(key, val);
  },

  _set: function(key, val) {
    var pos = this._nextIndex;
    this._keys[pos] = key;
    this._values[pos] = val;
    pos++;
    this._nextIndex = pos === this._keys.length ? 0 : pos;
  }

};
